#include "LscMonitor.hh"
#include "DVector.hh"
#include <string>

EXECDAT(LscMonitor)

using namespace std;

inline void
addCh(Dacc& d, const string& i, const char* c, TSeries** t) {
    string cname = i + ":LSC-" + c;
    *t   =   0;
    d.addChannel(cname.c_str(), 0, t);
}

//======================================  Monitor Constructor
LscMonitor::LscMonitor(int argc, const char* argv[]) 
  : DatEnv(argc, argv), mDynConf(getDacc()), mEmulator(mDynConf),
    mNSegment(0)
{
    string ifo;
    bool syntax = false;
    for (int i=1 ; i<argc ; i++) {
        string argi = argv[i];
	if (argi == "-ifo") {
	    ifo = argv[++i];
	} else if (isDatEnvArg(argv[i])) {
	    i++;
	} else {
	    cerr << "Error in command line syntax: " << argi 
		 << " is not understood." << endl;
	    syntax = true;
	}
    }
    if (ifo != "H1" && ifo != "H2" && ifo != "L1") {
        cerr << "The interferometer name was not specified or invalid" << endl;
	syntax = true;
    }
    if (syntax) {
        cerr << "Command syntax:" << endl; 
	cerr << argv[0] << " -ifo {H1 | H2 | L1} <DatEnv arguments>" << endl;
	finish();
	return;
    }

    //----------------------------------  Set up the emulator
    getDacc().setStride(16.0);
    getDacc().setIgnoreMissingChannel(true);
    mDynConf.setFile("LscEmul.param");
    mEmulator.setDebug(Debug());
    mEmulator.setIfo(ifo);

    //----------------------------------  Request data channels
    addCh(getDacc(), ifo, "AS_I"  , &mChanAsI);
    addCh(getDacc(), ifo, "AS_Q"  , &mChanAsQ);
    addCh(getDacc(), ifo, "POB_I" , &mChanPoI);
    addCh(getDacc(), ifo, "POB_Q" , &mChanPoQ);
    addCh(getDacc(), ifo, "REFL_I", &mChanRfI);
    addCh(getDacc(), ifo, "REFL_Q", &mChanRfQ);
    
    addCh(getDacc(), ifo, "DARM_CTRL", &mChanDArm);
    addCh(getDacc(), ifo, "MICH_CTRL", &mChanMich);
    addCh(getDacc(), ifo, "PRC_CTRL" , &mChanPrc);
    addCh(getDacc(), ifo, "CARM_CTRL", &mChanCArm);
}

//======================================  Monitor destructor
LscMonitor::~LscMonitor(void) {
}

//======================================  Monitor Constructor
void 
LscMonitor::ProcessData(void) {
    Time tStart = getDacc().getFillTime();
    Time tStop  = getDacc().getCurrentTime();
    Time tEnd   = tStop;
    for (Time t0=tStart ; t0 < tStop ; t0 = tEnd) {
        mEmulator.getLscConfig(t0);
	mEmulator.reset();
	tEnd = mDynConf.getExpiry();
	if (!tEnd) tEnd = tStop;
	Interval Dt = tEnd - t0;
	if (mChanAsI) mEmulator.setSeries(LscEmul::kDataAsI, 
					  mChanAsI->extract(t0,Dt));
	if (mChanAsQ) mEmulator.setSeries(LscEmul::kDataAsQ, 
					  mChanAsQ->extract(t0,Dt));
	if (mChanPoI) mEmulator.setSeries(LscEmul::kDataPoI, 
					  mChanPoI->extract(t0,Dt));
	if (mChanPoQ) mEmulator.setSeries(LscEmul::kDataPoQ, 
					  mChanPoQ->extract(t0,Dt));
	if (mChanRfI) mEmulator.setSeries(LscEmul::kDataRflI, 
					  mChanRfI->extract(t0,Dt));
	if (mChanRfQ) mEmulator.setSeries(LscEmul::kDataRflQ, 
					  mChanRfQ->extract(t0,Dt));
	if (mChanDArm) mEmulator.setSeries(LscEmul::kDArmCtl, 
					  mChanDArm->extract(t0,Dt));
	if (mChanMich) mEmulator.setSeries(LscEmul::kMichCtl, 
					  mChanMich->extract(t0,Dt));
	if (mChanPrc) mEmulator.setSeries(LscEmul::kPrcCtl, 
					  mChanPrc->extract(t0,Dt));
	if (mChanDArm) mEmulator.setSeries(LscEmul::kCArmCtl, 
					  mChanCArm->extract(t0,Dt));
	mEmulator.emulate();
	mNSegment++;

	//------------------------------  Compare the results
	if (mNSegment > 2) {
	    mEmulator.compare(LscEmul::kDArmCalc);
	    mEmulator.compare(LscEmul::kMichCalc);
	    mEmulator.compare(LscEmul::kPrcCalc);
	    mEmulator.compare(LscEmul::kCArmCalc);
	}
    }
}

//
//private:
//    LscEmul mEmulator;
