/* -*- mode: c++; c-basic-offset: 4; -*- */
/*
 * h(t) monitor - Generation of time-domain calibrated data.
 *
 * Jordi Burguet-Castell, Shourov K. Chatterji, Xavier Siemens.
 * Aug, Sep, Oct 2008.
 */

// $Id: hoftMon.cc,v 1.24 2009/08/19 22:50:50 jordi Exp $

#include "hoftMon.hh"


// The next three lines are needed if you are going to generate triggers.
// The text in PIDTITLE should describe the monitor function.
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/hoftMon/hoftMon.cc,v 1.24 2009/08/19 22:50:50 jordi Exp $"
#define PIDTITLE  "Generate calibrated h(t)"
#include "ProcIdent.hh"

#include "ParseLine.hh"

#include "Dacc.hh"
#include <iostream>
#include <string>
#include <cstdlib>

#if !defined(isinf) && defined(__sun__)
#include <ieeefp.h>
inline bool
isinf(double x) {
    return !finite(x);
}
#endif

extern "C" {
#include <lal/ReadFiltersFile.h>   // new file with Xavi's function
#include <lal/ComputeDataQualityVector.h>
#include <lal/AVFactories.h>       // to use XLALCreateREAL4Vector
}

using namespace std;


INT4 lalDebugLevel = 0;  /* for debugging calls to LAL functions, put it to 3 */

// Macro for log output.
#define LOG(LEVEL) if (Debug() >= LEVEL) cout << "hoftMon: "

// Macro for common checks.
#define CHECK(COND)                                                     \
    do {                                                                \
        if (COND != 0) {                                                \
            cerr << "Error in " << __FILE__ << ":" << __LINE__ << endl; \
            finish();  return;                                          \
        }                                                               \
    } while (0)


// Generate main routine.
EXECDAT(hoftMon)


/*
 * Constructor.
 */
hoftMon::hoftMon(int argc, const char *argv[])
  : DatEnv(argc, argv), mStride(16.0), mIteration(0), mStepsSinceLastDrop(0),
    mSVInChanName("Xn:IFO-SV_STATE_VECTOR"),
    mLAXChanName("Xn:LSC-LA_PTRX_NORM"), mLAYChanName("Xn:LSC-LA_PTRY_NORM"),
    mSVChanName("Xn:DMT-STATE_VECTOR"), mDQChanName("Xn:DMT-DATA_QUALITY_VECTOR"),
    mStrainChanName("Xn:DMT-STRAIN"), mOutputFileName("hoft-%s-16.gwf")
{
    int err = 0;  /* error code */
    
    LOG(3) << "Initializing\n";
    mInputData.Cinv = NULL;       // filters file still not read
    mInputData.DARM_ERR.data = NULL;  // lal vectors not allocated yet
    err = initialize_input_data(argc, argv);
    CHECK(err);
   
    // Initialize times of last bad times seen in previous blocks.
    m_t_bad_left_1 = m_t_bad_left_2 = m_t_bad_left_3 = 0;  // ugly, isn't it?
 
    // Set the data accessor.
    getDacc().setStride(mStride);  // we'll read chunks of mStride seconds
    for (int i = 0; i < 3; i++)
        getDacc().addChannel(mCalChanNames[i].c_str());  // add calibration

    getDacc().addChannel(mSVInChanName.c_str());  // add state vector
    getDacc().addChannel(mLAXChanName.c_str());   // add light-in-arms X
    getDacc().addChannel(mLAYChanName.c_str());   // add light-in-arms Y
}



/*
 * Destructor.
 */
hoftMon::~hoftMon() 
{
    if (mInputData.DARM_ERR.data != NULL)
        free_data();  // frees the LAL channels
    
    if (mInputData.Cinv != NULL)
        destroy_initialization();    // destroys filters file
    
    LOG(3) << "hoftMon is finished\n";
}


/*
 * Handle messages.
 */
void hoftMon::Attention(void)
{
    MonServer::Attention();
}



/* ------------------------- Main functions ---------------------------- */


/*
 * Process one data stride. This function gets called each time a new
 * frame is ready to be processed.
 */
void hoftMon::ProcessData()
{
    int err = 0;  /* error code */
    
    LOG(3) << "Starting ProcessData (iteration " << mIteration << ")\n";
    
    const TSeries* dmt_cal_chans[3];
    err = get_cal_channels(dmt_cal_chans);   // fills dmt_cal_chans
    CHECK(err);
    
    if (mIteration == 0) {
        LOG(3) << "Allocating memory for the LAL channels\n";
        err = allocate_data(dmt_cal_chans);  // allocate memory for LAL channels
        CHECK(err);
    }
    
    LOG(3) << "Reading data: fill LAL input and output structures\n";
    read_cal_data(dmt_cal_chans);  // fills LAL input and output structures

    long phase = long(dmt_cal_chans[0]->getStartTime().getS() / mStride);
    if ((phase % mNPhase) == mPhaseNo) {
        
    /* Compute calibrated h(t) by calling LALComputeStrain */
    if (mStepsSinceLastDrop >= 2) {
        LOG(3) << "Calling ComputeStrain\n";
        LALStatus status;
        status.statusPtr = NULL;   // same thing as "static LALStatus status;"
        LALComputeStrain(&status, &mOutputData, &mInputData);
    }

    /* Compute the last time that DQ up=0, so that we know when data
     * are properly calibrated */
    m_t_bad_left_3 = m_t_bad_left_2;  // keep the last bad times
    m_t_bad_left_2 = m_t_bad_left_1;
    TSeries dq = compute_dq(0, 0,
                            *getDacc().refData(mSVInChanName.c_str()),
                            *getDacc().refData(mLAXChanName.c_str()),
                            *getDacc().refData(mLAYChanName.c_str()));
    int* dq_data = (int*) dq.refData();
    int n = dq.getNSample();  // == 16
    int t_bad_right = n;
    for (int i = 0; i < n; i++)
        if ((dq_data[i] & (1 << 2)) == 0) {   // if not UP
            t_bad_right = i;
            break;
        }
    m_t_bad_left_1 = -n;
    for (int i = n-1; i >= 0; i--)
        if ((dq_data[i] & (1 << 2)) == 0) {   // if not UP
            m_t_bad_left_1 = -(n-1) + i;
            break;
        }

    if (mStepsSinceLastDrop >= 1) {
        mLastDQ = compute_dq(m_t_bad_left_3, t_bad_right,
                             mLastSV, mLastLAX, mLastLAY);
    }

    /* Write the frame */
    if (mStepsSinceLastDrop >= 1) {
        LOG(3) << "Writing the frame\n";
        err = write_frames(dmt_cal_chans);
        CHECK(err);
    }

    } // end of phase selection.

    mLastSV.copy(*getDacc().refData(mSVInChanName.c_str()));
    mLastLAX.copy(*getDacc().refData(mLAXChanName.c_str()));
    mLastLAY.copy(*getDacc().refData(mLAYChanName.c_str()));

    mIteration++;
    mStepsSinceLastDrop++;
    mLastTime = int(dmt_cal_chans[0]->getStartTime().getS() - 2*mStride);
}


/*
 * Fills mInputData with the times (epoch and deltaT) and channels,
 * extracted from the DMT. If it detects that there has been a
 * dropout, it also sets mStepsSinceLastDrop to 0.
 *
 * To be called in ProcessData().
 */
void hoftMon::read_cal_data(const TSeries* dmt_cal_chans[])
{
    /* Get epoch, deltaT's and numer of samples for each channel */
    
    // Get the start time of this chunk of data
    LIGOTimeGPS epoch;
    epoch.gpsSeconds     = (INT4) dmt_cal_chans[0]->getStartTime().getS()
                                                      - 2 * (INT4) mStride;
    epoch.gpsNanoSeconds = (INT4) dmt_cal_chans[0]->getStartTime().getN();
    // The 2*mStride (=32) seconds shift is done to get the epoch of
    // two strides of data ago
    
    // Check if there has been a dropout
    if (mIteration > 0 && epoch.gpsSeconds != mLastTime + mStride) {
        LOG(1) << "We lost some data! Times used for calibration:\n"
               << "  Last gps time:    " << mLastTime << endl
               << "  Current gps time: " << epoch.gpsSeconds << endl;
        mStepsSinceLastDrop = 0;
    }
    
    // Get the sampling period (deltaT) and the total number of samples
    REAL8 deltaT[3];
    TSeries::size_type nsamples[3];
    for (int i = 0; i < 3; i++) {
        deltaT[i] = (REAL8) dmt_cal_chans[i]->getTStep().GetSecs();
        nsamples[i] = dmt_cal_chans[i]->getNSample();
    }

    REAL8 To = mInputData.To;

    
    /* Input channels, copied from the dmt calibration channels */
    
    REAL4* chan_data[3] = {
//        mInputData.AS_Q.data->data, // see comment in hoftMon.hh
        mInputData.DARM.data->data,
        mInputData.DARM_ERR.data->data,
        mInputData.EXC.data->data
    };
    
    for (int i = 0; i < 3; i++) {
        int n = (int) nsamples[i];
        
        // Shift all data to the left.
        for (int j = 0; j < 2*n; j++)
            chan_data[i][j] = chan_data[i][j+n];
        
        // Write new data in the last part of the array.
        dmt_cal_chans[i]->getData(nsamples[i], chan_data[i] + 2 * nsamples[i]);
    }


    /* Fill times (epoch and deltaT) */

    mInputData.AS_Q.epoch        = epoch;
    mInputData.DARM.epoch        = epoch;
    mInputData.DARM_ERR.epoch    = epoch;
    mInputData.EXC.epoch         = epoch;
    mOutputData.h.epoch          = epoch;
    mOutputData.hC.epoch         = epoch;
    mOutputData.hR.epoch         = epoch;
    mOutputData.alpha.epoch      = epoch;
    mOutputData.beta.epoch       = epoch;
    mOutputData.alphabeta.epoch  = epoch;

    mInputData.AS_Q.deltaT       = deltaT[0];
    mInputData.DARM.deltaT       = deltaT[0];
    mInputData.DARM_ERR.deltaT   = deltaT[1];
    mInputData.EXC.deltaT        = deltaT[2];
    mOutputData.h.deltaT         = deltaT[1];  // same as DARM_ERR
    mOutputData.hC.deltaT        = deltaT[1];  // same as DARM_ERR
    mOutputData.hR.deltaT        = deltaT[1];  // same as DARM_ERR
    mOutputData.alpha.deltaT     = To;
    mOutputData.beta.deltaT      = To;
    mOutputData.alphabeta.deltaT = To;
}


/*
 * Write the calibrated h(t) frames to disk.
 *
 * mOutputData must have the channel h already filled (by calling
 * LALComputeStrain at some point before this function is called).
 *
 * The dmt_cal_chans is passed just to get the start time and time step.
 */
int hoftMon::write_frames(const TSeries* dmt_cal_chans[])
{
    /*
     * Extract things from the LAL structures and put it back into
     * something readable by the DMT.
     */
    UINT4 length = mOutputData.h.data->length / 3;  // take just the middle 1/3

    TSeries hoft;

    hoft.setName("Calibrated h(t)");

    // Copy only the middle segment into hoft
    hoft.setData(dmt_cal_chans[0]->getStartTime() - Interval(mStride),
                 dmt_cal_chans[0]->getTStep(),
                 new DVecType<double>(length,
                                      mOutputData.h.data->data + length));
    // The "new" is necessary, because setData destroys the content of
    // the vector that is passed as argument (through the use of a
    // auto_ptr<> that makes a reset())
    
    /*
     * Write it all.
     */
    //mFrameWriter.setCompress(FrVectRef::kZeroSuppOrGZip);
    mFrameWriter.buildFrame(hoft.getStartTime(), Interval(mStride));

    mFrameWriter.addProcSeries(mStrainChanName.c_str(), hoft); // h(t)
    mFrameWriter.addProcSeries(mSVChanName.c_str(), mLastSV);  // state vector
    mFrameWriter.addProcSeries(mDQChanName.c_str(), mLastDQ);  // data quality
    
    mFrameWriter.addHistory("hoftMon", Now(), "online h(t) calibrated data");
    mFrameWriter.addHistory("hoftMon", Now(), mHeader);
    mFrameWriter.addWriterHistory();

    char fname[256];  // buffer for the file name
    TimeStr(dmt_cal_chans[0]->getStartTime() - Interval(16),
            fname, mOutputFileName.c_str());  // allow a nice name for the file

    mFrameWriter.open(fname, true);
    mFrameWriter.writeFrame();
    mFrameWriter.close();

    return 0;
}




/* --------------------- Auxiliary functions ---------------------------- */


/*
 * Get pointers to the contents of the different calibration channels.
 */
int hoftMon::get_cal_channels(const TSeries* dmt_cal_chans[])
{
    // Get the channels by calling getDacc()
    for (int i=0; i < 3; i++)
        dmt_cal_chans[i] = getDacc().refData(mCalChanNames[i].c_str());
    
    // Check that all channels are there indeed.
    for (int i=0; i < 3; i++)
        if (!(dmt_cal_chans[i]) || (dmt_cal_chans[i])->isEmpty()) {
            cerr << "Channel " << mCalChanNames[i] << " was not found.\n";
            return -1;
        }
    
    return 0;
}


/*
 * Set the names of all channels to their default values with the name
 * of the interferometer as its initial part.
 */
void hoftMon::set_channel_names(string ifo)
{
    /* Check that the IFO name is 2 letters long */
    if (ifo.length() != 2) {
        cerr << "Warning: IFO name (" << ifo << ") doesn't have "
             <<" just 2 characters! I won't change the channel names.\n";
        return;
    }
    
    /* Change name of input channels */
    for (int i = 0; i < 3; i++)  // calibration channels
        mCalChanNames[i] = ifo + mCalChanNames[i].substr(2);

    mSVInChanName = ifo + mSVInChanName.substr(2);
    mLAXChanName  = ifo + mLAXChanName.substr(2);
    mLAYChanName  = ifo + mLAYChanName.substr(2);

    /* Change name of output channels */
    mSVChanName      = ifo + mSVChanName.substr(2);
    mDQChanName      = ifo + mDQChanName.substr(2);
    mStrainChanName  = ifo + mStrainChanName.substr(2);
}



/*
 * Compute the Data Quality channel from the given State Vector,
 * Light-in-x-arm and Light-in-y-arm series. The t_bad_left and
 * t_bad_right are the (signed) number of seconds when the last bad
 * data (detector not UP) happened, with respect to this frame.
 */
TSeries hoftMon::compute_dq(int t_bad_left, int t_bad_right,
                            TSeries& SV, TSeries& LAX, TSeries& LAY)
{
    /* Define the DQ series time step (1/rate) */
    const float dq_step = 1.0;

    /* Get what I need from the LAX, LAY and SV channels */
    float* sv_data = (float*) SV.refData();
    int r_sv = int(dq_step / SV.getTStep());
    // rate: number of state vector data per data quality step

    float* lax_data = (float*) LAX.refData();
    float* lay_data = (float*) LAY.refData();
    int r_light = int(dq_step / LAX.getTStep());

    /* Array with the contents of the DQ vector */
    int n_dq = int(mStride / dq_step);
    int* dq_data = new int[n_dq];

    /* Short name for alphabeta */
    COMPLEX16Sequence* gamma = mOutputData.alphabeta.data;
    UINT4 n_gamma = gamma->length / 3;  // we want just the middle 1/3

    /* End touch: call LAL and put the resulting dq_data in a nice TSeries */
    XLALComputeDQ(sv_data, r_sv,
                  lax_data, lay_data, r_light,
                  gamma->data + n_gamma, n_gamma/n_dq,  // gamma_data shifted
                  t_bad_left, t_bad_right, mInputData.wings,
                  0,
                  dq_data, n_dq);

    TSeries dq(SV.getStartTime(), Interval(dq_step),
               int(mStride/dq_step), dq_data);

    dq.setName("Data Quality vector");

//    delete[] dq_data;  //// ?  maybe the TSeries last_dq already deletes it??

    return dq;
}



/* ------------------ Initialization, allocation et al ------------------- */


/*
 * Read command line arguments and initialize internal variables from
 * it. Set all the non-changing input variables (arguments passed to
 * the program). To be called in the constructor.
 */
int hoftMon::initialize_input_data(int argc, const char *argv[])
{
    int err = 0;  /* error code */
    const char* filters_file = NULL;  /* temporary pointer to name of the filters file */
    
    /* Initialize default values (fill all but the time series) */
    
    mInputData.Go.re = -0.44761314397906;  // open loop gain at cal line freq (real)
    mInputData.Go.im = -0.09999284928620;  // (imaginary part)
    mInputData.Do.re = -0.14346135865327;  // digital filter at cal line freq (real)
    mInputData.Do.im =  0.22204243370711;  // (imaginary part)
    mInputData.Wo.re =  0.00987795632736;  // whitening filter at cal line freq (real)
    mInputData.Wo.im = -0.00156482799222;  // (imaginary part)
    mInputData.f     = 407.3;  // frequency of the calibration line
    mInputData.To    = 1.0;    // factors integration time (in s)
    // These default values come from
    // ~xsiemens/hoft/S5RecoveryApril2007/H2/DONEPhase2/S5strainDAGH2_822760000-824862720.ini
    // at ldas-pcdev1.ligo.caltech.edu.
    // They should be given as parameters passed when calling the program.

    mInputData.delta      = 0;
    mInputData.wings      = 16;  // width (in s) of the "extra" data for the calibration
    mInputData.usefactors = 1;
    mInputData.fftconv    = 1;
    mInputData.outalphas  = 0;
    mInputData.darmctrl   = 1;
    mInputData.gamma_fudgefactor = 1;

    /* Initialize phase data */
    mNPhase  = 1;
    mPhaseNo = 0;

    /* Read data from the command line. */

    char usage[] =
"Command syntax:\n"
"hoftMon [-help] [-debug <level>]"
" [-inlist <value>]"
" [-conf <value>]"
" [-stride <value>]"
" [-ifo <value>]"
" [-o <value>]"
" [-cal-line-freq <freq>]"
" [-olg-re <value>] [-olg-im <value>]"
" [-servo-re <value>] [-servo-im <value>]"
" [-whitener-re <value>] [-whitener-im <value>]"
" [-factors-time <time>]"
" [-wings <value>]"
" [-gamma-fudge-factor <value>]"
" [-phase <i>:<N>]"
" [-delta] [-td-fir] [-no-factors] [<DatEnv-args>]\n";

    char help[] =
"Arguments are (a * means that they are not yet implemented):\n"
"  -conf                STRING  Config file with parameter values.\n"
"  -stride              FLOAT   Length in seconds of the data stride read from the DMT.\n"
"  -ifo                 STRING  Interferometer name (eg, L1).\n"
"  -o                   STRING  Output file name (eg, hoft-%s-16.gwf).\n"
"  -cal-line-freq       FLOAT   Calibration line frequency in Hz.\n"
"* -factors-time        FLOAT   Factors integration time in seconds.\n"
"  -olg-re              FLOAT   Real part of the open loop gain at the calibration line frequency.\n"
"  -olg-im              FLOAT   Imaginary part of the open loop gain at the calibration line frequency.\n"
"  -servo-re            FLOAT   Real part of the digital filter at the calibration line frequency.\n"
"  -servo-im            FLOAT   Imaginary part of digital filter at the calibration line frequency.\n"
"  -whitener-re         FLOAT   Real part of the whitening filter at the calibration line frequency.\n"
"  -whitener-im         FLOAT   Imaginary part of whitening filter at the calibration line frequency.\n"
"  -filters-file        STRING  Name of file containing filters and histories.\n"
"  -wings               INT     Size of wings in seconds.\n"
"* -test-sensing        FLAG    Output residual strain only.\n"
"* -test-actuation      FLAG    Output control strain only.\n"
"  -delta               FLAG    Use unit impulse.\n"
"  -no-factors          FLAG    Do not use factors in strain computation.\n"
"  -td-fir              FLAG    Use time-domain FIR filtering (default is FFT convolution).\n"
"  -gamma-fudge-factor  FLOAT   Fudge factor used to adjust factor values. Gamma is divided by that value.\n"
"* -frame-type          STRING  Frame type to be written (eg, H1_RDS_C01_LX).\n"
"  -debug               INT     Debug level (cannot be in config file).\n"
"  -inlist              STRING  File with the paths to the input frame files (cannot be in config file).\n"
"  -version             FLAG    Show version number and exit.\n"
"  -help                FLAG    This message.\n";
    
    // Look for arguments.
    for (int i = 1; i < argc; i++) {
        string s = argv[i];
        
        if (s == "-help") {
            cout << usage << endl << help;
            finish();  return 0;
        }
        else if (s == "-version") {
            cout << argv[0] << " version 0.98" << endl;
            finish();  return 0;
        }
        else if (s == "-stride")          { mStride = strtod(argv[++i], 0); }
        else if (s == "-ifo")             { set_channel_names(argv[++i]); }
        else if (s == "-o")               { mOutputFileName = argv[++i]; }
        else if (s == "-no-factors")      { mInputData.usefactors = 0; }
        else if (s == "-olg-re")          { mInputData.Go.re = strtod(argv[++i], 0); }
        else if (s == "-olg-im")          { mInputData.Go.im = strtod(argv[++i], 0); }
        else if (s == "-servo-re")        { mInputData.Do.re = strtod(argv[++i], 0); }
        else if (s == "-servo-im")        { mInputData.Do.im = strtod(argv[++i], 0); }
        else if (s == "-whitener-re")     { mInputData.Wo.re = strtod(argv[++i], 0); }
        else if (s == "-whitener-im")     { mInputData.Wo.im = strtod(argv[++i], 0); }
        else if (s == "-cal-line-freq")   { mInputData.f = strtod(argv[++i], 0); }
        else if (s == "-factors-time")    { mInputData.To = strtod(argv[++i], 0); }
        else if (s == "-delta")           { mInputData.delta = 1; }
        else if (s == "-wings")           { mInputData.wings = atoi(argv[++i]); }
        else if (s == "-td-fir")          { mInputData.fftconv = 0; }
        else if (s == "-gamma-fudge-factor") { mInputData.gamma_fudgefactor = strtod(argv[++i], 0); }
        else if (s == "-filters-file") {
            filters_file = argv[++i];
            err = XLALReadFiltersFile(filters_file, &mInputData);
            if (err)
                return err;
        }
        else if (s == "-conf") {
            err = read_config_file(argv[++i]);
            if (err)
                return err;
        }
        else if (s == "-phase") {
            char* p = const_cast<char*>(argv[++i]);
            mPhaseNo = strtol(p, &p, 0);
            if (*p++ != ':') return 22;
            mNPhase = strtol(p, &p, 0);
        }
        else if (isDatEnvArg(argv[i])) {
            i++;
        }
        else {
            cerr << "Unrecognized argument: " << s << endl
                 << usage;
            return -3;
        }
    }
    
    /* Check that the duration makes sense (is long enough) */
    
    if (mStride < 10.0) {  // actuation filter takes 10 seconds
        cerr << "Time stride is too small (" << mStride << " s) "
             << "No way I can calibrate this!\n";
        return -1;
    }
    else if (fabs(mStride - 16.0) > 0.01) {  // nothing special about the 0.01
        cerr << "Warning: duration (" << mStride << " s) "
             << "different from 16 s. I'll continue, though.\n";
    }

    /* Get command line and all parameters as a string for the header */

    mHeader += "Command line run: ";
    for (int i = 0; i < argc; i++) {
        mHeader += argv[i];
        mHeader += " ";
    }

    char buffer[1000];
    snprintf(buffer, 1000, "\nActual parameters:  "
             "stride=%.15e  usefactors=%d  Go.re=%.15e  Go.im=%.15e  "
             "Do.re=%.15e  Do.im=%.15e  Wo.re=%.15e  Wo.im=%.15e  "
             "f=%.15e  To=%.15e  delta=%d  wings=%d  fftconv=%d  "
             "gamma_fudgefactor=%.15e  filters_file=%s",
             mStride, mInputData.usefactors, mInputData.Go.re, mInputData.Go.im,
             mInputData.Do.re, mInputData.Do.im, mInputData.Wo.re, mInputData.Wo.im,
             mInputData.f, mInputData.To, mInputData.delta, mInputData.wings, mInputData.fftconv,
             mInputData.gamma_fudgefactor, filters_file);
    mHeader += buffer;

    setStrideAlignment(long(mStride), 0.0);

    return 0;
}



/*
 * Read parameters from a file and initialize internal variables from
 * it. Called from initialize_input_data() if option -conf is used.
 */
int hoftMon::read_config_file(const char* cfile)
{
    int err;

    ParseLine pl(cfile);
    if (! pl.isOpen()) {
        LOG(1) << "Unable to open configuration file: " << cfile << endl;
        return -10;
    }

    while (pl.getLine() > 0) {  // number of arguments or negative error
        string s = pl[0];

        if      (s == "stride")          { mStride = strtod(pl[1], 0); }
        else if (s == "ifo")             { set_channel_names(pl[1]); }
        else if (s == "o")               { mOutputFileName = pl[1]; }
        else if (s == "no-factors")      { mInputData.usefactors = 0; }
        else if (s == "olg-re")          { mInputData.Go.re = strtod(pl[1], 0); }
        else if (s == "olg-im")          { mInputData.Go.im = strtod(pl[1], 0); }
        else if (s == "servo-re")        { mInputData.Do.re = strtod(pl[1], 0); }
        else if (s == "servo-im")        { mInputData.Do.im = strtod(pl[1], 0); }
        else if (s == "whitener-re")     { mInputData.Wo.re = strtod(pl[1], 0); }
        else if (s == "whitener-im")     { mInputData.Wo.im = strtod(pl[1], 0); }
        else if (s == "cal-line-freq")   { mInputData.f = strtod(pl[1], 0); }
        else if (s == "factors-time")    { mInputData.To = strtod(pl[1], 0); }
        else if (s == "delta")           { mInputData.delta = 1; }
        else if (s == "wings")           { mInputData.wings = atoi(pl[1]); }
        else if (s == "td-fir")          { mInputData.fftconv = 0; }
        else if (s == "gamma-fudge-factor") { mInputData.gamma_fudgefactor = strtod(pl[1], 0); }
        else if (s == "filters-file") {
            err = XLALReadFiltersFile(pl[1], &mInputData);
            if (err)
                return err;
        }
        else if (s == "phase") {
            char* p = const_cast<char*>(pl[1]);
            mPhaseNo = strtol(p, &p, 0);
            if (*p++ != ':') return 22;
            mNPhase = strtol(p, &p, 0);
        }
        else if (isDatEnvArg(("-" + s).c_str())) {
            continue;
        }
        else {
            cerr << "Unrecognized argument: " << s << endl;
            return -30;
        }
    }
    
    return 0;
}


/*
 * Allocate space looking at how many samples there are in the dmt
 * channels (dmt_cal_chans).
 */
int hoftMon::allocate_data(const TSeries* dmt_cal_chans[])
{
    /* Find the size of the vectors we will need to allocate */
    
    UINT4 nsamples[3];
    
    // All samples have the same duration (so we get it from the first)
    REAL8 duration = (REAL8) dmt_cal_chans[0]->getInterval().GetSecs();
    
    // Number of samples = duration / deltaT
    for (int i=0; i < 3; i++) {
        REAL8 deltaT = (REAL8) dmt_cal_chans[i]->getTStep().GetSecs();
        nsamples[i]  = (UINT4)(duration / deltaT + 0.5);
    }

    // The output data has the same size as DARM_ERR
    UINT4 nsampl_h = nsamples[1];
    
    // The alpha et al have a different deltaT
    REAL8 To = mInputData.To;
    UINT4 nsampl_alpha = (UINT4)(duration / To + 0.5);
    
    /* Check that it makes sense */
    
    // nsamples as expected from the dmt channels?
    for (int i = 0; i < 3; i++) {
        UINT4 n_dmt = (UINT4) dmt_cal_chans[i]->getNSample();
        if (nsamples[i] != n_dmt) {
            cerr << "Error. Durations differ for channel " << mCalChanNames[i]
                 << "\nDMT: " << n_dmt << "\tComputed: " << nsamples[i] << endl;
            return -1;
        }
    }
    
    /* Allocate space for vectors in LAL input data */
    
    // The 3* factor accounts for the buffering of data.
    mInputData.AS_Q.data     = XLALCreateREAL4Vector(3 * nsamples[0]);
    mInputData.DARM.data     = XLALCreateREAL4Vector(3 * nsamples[0]);
    mInputData.DARM_ERR.data = XLALCreateREAL4Vector(3 * nsamples[1]);
    mInputData.EXC.data      = XLALCreateREAL4Vector(3 * nsamples[2]);
    if (!mInputData.DARM.data || !mInputData.DARM_ERR.data || !mInputData.EXC.data) {
        cerr << "Error allocating space for input channels.\n";
        return -2;
    }

    /* Allocate space for vectors in LAL output data */
    
    mOutputData.h.data  = XLALCreateREAL8Vector(3 * nsampl_h);
    mOutputData.hC.data = XLALCreateREAL8Vector(3 * nsampl_h);
    mOutputData.hR.data = XLALCreateREAL8Vector(3 * nsampl_h);
    
    mOutputData.alpha.data     = XLALCreateCOMPLEX16Vector(3 * nsampl_alpha);
    mOutputData.beta.data      = XLALCreateCOMPLEX16Vector(3 * nsampl_alpha);
    mOutputData.alphabeta.data = XLALCreateCOMPLEX16Vector(3 * nsampl_alpha);
    
    if (!mOutputData.h.data || !mOutputData.hC.data || !mOutputData.hR.data ||
        !mOutputData.alpha.data || !mOutputData.beta.data || !mOutputData.alphabeta.data) {
        cerr << "Error allocating channel vector.\n";
        return -4;
    }
    
    return 0;
}



/*
 * Free the allocated memory.
 */
int hoftMon::free_data()
{
    /* Free input data */
    XLALDestroyREAL4Vector(mInputData.AS_Q.data);
    XLALDestroyREAL4Vector(mInputData.DARM.data);
    XLALDestroyREAL4Vector(mInputData.DARM_ERR.data);
    XLALDestroyREAL4Vector(mInputData.EXC.data);      

    /* Free output data. */
    XLALDestroyREAL8Vector(mOutputData.h.data);
    XLALDestroyREAL8Vector(mOutputData.hC.data);
    XLALDestroyREAL8Vector(mOutputData.hR.data);
    XLALDestroyCOMPLEX16Vector(mOutputData.alpha.data);
    XLALDestroyCOMPLEX16Vector(mOutputData.beta.data);
    XLALDestroyCOMPLEX16Vector(mOutputData.alphabeta.data);

    return 0;
}



/*
 * Destroys the extra objects created at initialize_data(). To be
 * called in the destructor.
 */
int hoftMon::destroy_initialization()
{
    int err = XLALDestroyFiltersFile(&mInputData);
    if (err) {
        cerr << "Error destroying filters file.\n";
        return -1;
    }

    return 0;
}
